"""This module deals with generating transactions."""
from typing import Any, Protocol

from solana.transaction import Transaction

from anchorpy.program.context import EMPTY_CONTEXT, Context, _check_args_length
from anchorpy.idl import _IdlInstruction
from anchorpy.program.namespace.instruction import _InstructionFn


class _TransactionFn(Protocol):
    """A function to create a `Transaction` for a given program instruction."""

    def __call__(self, *args: Any, ctx: Context = EMPTY_CONTEXT) -> Transaction:
        """Make sure that the function looks like this.

        Args:
            *args: The positional arguments for the program. The type and number
                of these arguments depend on the program being used.
            ctx: non-argument parameters to pass to the method.

        """
        ...


# ts TransactionNamespaceFactory.build
def _build_transaction_fn(
    idl_ix: _IdlInstruction, ix_fn: _InstructionFn
) -> _TransactionFn:
    """Build the function that generates Transaction objects.

    Args:
        idl_ix: Instruction item from the IDL.
        ix_fn (_InstructionFn): The function that generates instructions.

    Returns:
        _TransactionFn: [description]
    """

    def tx_fn(*args: Any, ctx: Context = EMPTY_CONTEXT) -> Transaction:
        _check_args_length(idl_ix, args)
        tx = Transaction()
        if ctx.pre_instructions:
            tx.add(*ctx.pre_instructions)
        tx.add(ix_fn(*args, ctx=ctx))
        if ctx.post_instructions:
            tx.add(*ctx.post_instructions)
        return tx

    return tx_fn



"""This module provides `AccountsCoder` and `_account_discriminator`."""
from hashlib import sha256
from typing import Tuple, Any
from construct import Adapter, Sequence, Bytes, Switch, Container

from anchorpy.coder.idl import _typedef_layout
from anchorpy.idl import Idl
from anchorpy.program.common import Instruction as AccountToSerialize

ACCOUNT_DISCRIMINATOR_SIZE = 8  # bytes


class AccountsCoder(Adapter):
    """Encodes and decodes account data."""

    def __init__(self, idl: Idl) -> None:
        """Init.

        Args:
            idl: The parsed IDL object.
        """
        self._accounts_layout = {
            acc.name: _typedef_layout(acc, idl.types, acc.name) for acc in idl.accounts
        }
        self.acc_name_to_discriminator = {
            acc.name: _account_discriminator(acc.name) for acc in idl.accounts
        }
        self.discriminator_to_acc_name = {
            disc: acc_name for acc_name, disc in self.acc_name_to_discriminator.items()
        }
        discriminator_to_typedef_layout = {
            disc: self._accounts_layout[acc_name]
            for acc_name, disc in self.acc_name_to_discriminator.items()
        }
        subcon = Sequence(
            "discriminator" / Bytes(ACCOUNT_DISCRIMINATOR_SIZE),
            Switch(lambda this: this.discriminator, discriminator_to_typedef_layout),
        )
        super().__init__(subcon)  # type: ignore

    def decode(self, obj: bytes) -> Container[Any]:
        """Decode account data.

        Args:
            obj: Data to decode.

        Returns:
            Decoded data.
        """
        return self.parse(obj).data

    def _decode(self, obj: Tuple[bytes, Any], context, path) -> AccountToSerialize:
        return AccountToSerialize(
            data=obj[1],
            name=self.discriminator_to_acc_name[obj[0]],
        )

    def _encode(self, obj: AccountToSerialize, context, path) -> Tuple[bytes, Any]:
        discriminator = self.acc_name_to_discriminator[obj.name]
        return discriminator, obj.data


def _account_discriminator(name: str) -> bytes:
    """Calculate unique 8 byte discriminator prepended to all anchor accounts.

    Args:
        name: The account name.

    Returns:
        The discriminator in bytes.
    """
    return sha256(f"account:{name}".encode()).digest()[:ACCOUNT_DISCRIMINATOR_SIZE]



"""IDL coding."""
from dataclasses import make_dataclass, asdict, fields as dc_fields
from types import MappingProxyType
from keyword import kwlist
from typing import Mapping, Optional, cast, Type, Sequence
from solana.publickey import PublicKey

from construct import Construct
from borsh_construct import (
    CStruct,
    TupleStruct,
    Vec,
    Enum,
    Bool,
    U8,
    I8,
    U16,
    I16,
    U32,
    I32,
    U64,
    I64,
    U128,
    I128,
    Bytes,
    String,
    Option,
)

from anchorpy.borsh_extension import _BorshPubkey, _DataclassStruct
from anchorpy.idl import (
    _IdlEnumFieldsNamed,
    _IdlEnumFieldsTuple,
    _IdlField,
    _IdlType,
    _IdlTypeArray,
    _IdlTypeDef,
    _IdlTypeDefTyEnum,
    _IdlTypeDefTyStruct,
    _IdlTypeDefined,
    _IdlTypeOption,
    _IdlTypeCOption,
    _IdlTypeVec,
    _NonLiteralIdlTypes,
)


FIELD_TYPE_MAP: Mapping[str, Construct] = MappingProxyType(
    {
        "bool": Bool,
        "u8": U8,
        "i8": I8,
        "u16": U16,
        "i16": I16,
        "u32": U32,
        "i32": I32,
        "u64": U64,
        "i64": I64,
        "u128": U128,
        "i128": I128,
        "bytes": Bytes,
        "string": String,
        "publicKey": _BorshPubkey,
    },
)


FIELD_PYTHON_TYPE_MAP: Mapping[str, Type] = MappingProxyType(
    {
        "bool": bool,
        "u8": int,
        "i8": int,
        "u16": int,
        "i16": int,
        "u32": int,
        "i32": int,
        "u64": int,
        "i64": int,
        "u128": int,
        "i128": int,
        "bytes": bytes,
        "string": str,
        "publicKey": PublicKey,
    },
)

_enums_cache: dict[tuple[str, str], Enum] = {}


def _handle_enum_variants(
    idl_enum: _IdlTypeDefTyEnum,
    types: Sequence[_IdlTypeDef],
    name: str,
) -> Enum:
    dict_key = (name, str(idl_enum))
    try:
        return _enums_cache[dict_key]
    except KeyError:
        result = _handle_enum_variants_no_cache(idl_enum, types, name)
        _enums_cache[dict_key] = result
        return result


def _handle_enum_variants_no_cache(
    idl_enum: _IdlTypeDefTyEnum,
    types: Sequence[_IdlTypeDef],
    name: str,
) -> Enum:
    variants = []
    dclasses = {}
    tuples = {}
    for variant in idl_enum.variants:
        variant_name = variant.name
        if variant.fields is None:
            variants.append(variant_name)
        else:
            variant_fields = variant.fields
            if isinstance(variant_fields[0], _IdlField):
                fields = []
                named_fields = cast(_IdlEnumFieldsNamed, variant_fields)
                for fld in named_fields:
                    fields.append(_field_layout(fld, types))
                cstruct = CStruct(*fields)
                datacls = _idl_enum_fields_named_to_dataclass_type(
                    named_fields,
                    types,
                    variant_name,
                )
                dclasses[variant_name] = datacls
                renamed = variant_name / cstruct
            else:
                fields = []
                unnamed_fields = cast(_IdlEnumFieldsTuple, variant_fields)
                for type_ in unnamed_fields:
                    fields.append(_type_layout(type_, types))
                tuple_struct = TupleStruct(*fields)
                tuple_ = _idl_enum_fields_tuple_to_tuple_type(unnamed_fields, types)
                tuples[variant_name] = tuple_
                renamed = variant_name / tuple_struct
            variants.append(renamed)  # type: ignore
    enum_without_types = Enum(*variants, enum_name=name)
    if dclasses:
        for cname in enum_without_types.enum._sumtype_constructor_names:
            try:
                dclass = dclasses[cname]
            except KeyError:
                continue
            dclass_fields = dc_fields(dclass)
            constructr = getattr(enum_without_types.enum, cname)
            for constructor_field in constructr._sumtype_attribs:
                attrib = constructor_field[1]  # type: ignore
                fld_name = constructor_field[0]  # type: ignore
                dclass_field = [f for f in dclass_fields if f.name == fld_name][0]
                attrib.type = dclass_field.type  # type: ignore
    if tuples:
        for cname_ in enum_without_types.enum._sumtype_constructor_names:
            try:
                tup = tuples[cname_]
            except KeyError:
                continue
            constructr = getattr(enum_without_types.enum, cname)
            constructr._sumtype_attribs[0][1].type = tup  # type: ignore
    return enum_without_types


def _typedef_layout_without_field_name(
    typedef: _IdlTypeDef,
    types: Sequence[_IdlTypeDef],
) -> Construct:
    typedef_type = typedef.type
    name = typedef.name
    if isinstance(typedef_type, _IdlTypeDefTyStruct):
        field_layouts = [_field_layout(field, types) for field in typedef_type.fields]
        cstruct = CStruct(*field_layouts)
        datacls = _idl_typedef_ty_struct_to_dataclass_type(typedef_type, types, name)
        return _DataclassStruct(cstruct, datacls=datacls)
    elif isinstance(typedef_type, _IdlTypeDefTyEnum):
        return _handle_enum_variants(typedef_type, types, name)
    unknown_type = typedef_type.kind
    raise ValueError(f"Unknown type {unknown_type}")


def _typedef_layout(
    typedef: _IdlTypeDef, types: Sequence[_IdlTypeDef], field_name: str
) -> Construct:
    """Map an IDL typedef to a `Construct` object.

    Args:
        typedef: The IDL typedef object.
        types: IDL type definitions.
        field_name: The name of the field.

    Raises:
        ValueError: If an unknown type is passed.

    Returns:
        `Construct` object from `borsh-construct`.
    """  # noqa: DAR402
    return field_name / _typedef_layout_without_field_name(typedef, types)


def _type_layout(type_: _IdlType, types: Sequence[_IdlTypeDef]) -> Construct:
    if isinstance(type_, str):
        return FIELD_TYPE_MAP[type_]
    field_type = cast(
        _NonLiteralIdlTypes,
        type_,
    )
    if isinstance(field_type, _IdlTypeVec):
        return Vec(_type_layout(field_type.vec, types))
    elif isinstance(field_type, _IdlTypeOption):
        return Option(_type_layout(field_type.option, types))
    elif isinstance(field_type, _IdlTypeCOption):
        raise NotImplementedError("COption not yet supported.")
    elif isinstance(field_type, _IdlTypeDefined):
        defined = field_type.defined
        if not types:
            raise ValueError("User defined types not provided")
        filtered = [t for t in types if t.name == defined]
        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        return _typedef_layout_without_field_name(filtered[0], types)
    elif isinstance(field_type, _IdlTypeArray):
        array_ty = field_type.array[0]
        array_len = field_type.array[1]
        inner_layout = _type_layout(array_ty, types)
        return inner_layout[array_len]
    raise ValueError(f"Type {field_type} not implemented yet")


def _field_layout(field: _IdlField, types: Sequence[_IdlTypeDef]) -> Construct:
    """Map IDL spec to `borsh-construct` types.

    Args:
        field: field object from the IDL.
        types: IDL type definitions.

    Raises:
        ValueError: If the user-defined types are not provided.
        ValueError: If the type is not found.
        ValueError: If the type is not implemented yet.

    Returns:
        `Construct` object from `borsh-construct`.
    """  # noqa: DAR402
    field_name = field.name if field.name else ""
    return field_name / _type_layout(field.type, types)


def _idl_type_to_python_type(
    idl_type: _IdlType,
    types: Sequence[_IdlTypeDef],
) -> Type:
    """Find the Python type corresponding to an IDL type.

    Args:
        idl_type: The IDL type.
        types: IDL type definitions.

    Raises:
        ValueError: If the user-defined types are not provided.
        ValueError: If the user-defined type is not found.

    Returns:
        The Python type.
    """
    if isinstance(idl_type, str):
        return FIELD_PYTHON_TYPE_MAP[idl_type]
    compound_idl_type = cast(
        _NonLiteralIdlTypes,
        idl_type,
    )
    if isinstance(compound_idl_type, _IdlTypeVec):
        type_arg = _idl_type_to_python_type(compound_idl_type.vec, types)
        return list[type_arg]  # type: ignore
    elif isinstance(compound_idl_type, _IdlTypeOption):
        return Optional[  # type: ignore
            _idl_type_to_python_type(compound_idl_type.option, types)
        ]
    elif isinstance(compound_idl_type, _IdlTypeArray):
        array_ty = compound_idl_type.array[0]
        array_len = compound_idl_type.array[1]
        return tuple[  # type: ignore
            (
                _idl_type_to_python_type(
                    array_ty,
                    types,
                ),
            )
            * array_len
        ]
    elif isinstance(compound_idl_type, _IdlTypeDefined):
        defined = compound_idl_type.defined
        if not types:
            raise ValueError("User defined types not provided")
        filtered = [t for t in types if t.name == defined]
        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        return _idl_typedef_to_python_type(
            filtered[0],
            types,
        )
    raise ValueError(f"Unrecognised type: {idl_type}")


def _datacls_cmp(left, right) -> bool:
    return (
        asdict(left) == asdict(right)
        and left.__class__.__name__ == right.__class__.__name__
    )


def _make_datacls(name: str, fields: list[tuple[str, type]]) -> type:
    return make_dataclass(name, fields, namespace={"__eq__": _datacls_cmp})


_idl_typedef_ty_struct_to_dataclass_type_cache: dict[tuple[str, str], Type] = {}


def _idl_typedef_ty_struct_to_dataclass_type(
    typedef_type: _IdlTypeDefTyStruct,
    types: Sequence[_IdlTypeDef],
    name: str,
) -> Type:
    dict_key = (name, str(typedef_type))
    try:
        return _idl_typedef_ty_struct_to_dataclass_type_cache[dict_key]
    except KeyError:
        result = _idl_typedef_ty_struct_to_dataclass_type_no_cache(
            typedef_type, types, name
        )
        _idl_typedef_ty_struct_to_dataclass_type_cache[dict_key] = result
        return result


def _idl_typedef_ty_struct_to_dataclass_type_no_cache(
    typedef_type: _IdlTypeDefTyStruct,
    types: Sequence[_IdlTypeDef],
    name: str,
) -> Type:
    """Generate a dataclass definition from an IDL struct.

    Args:
        typedef_type: The IDL type.
        types: IDL type definitions.
        name: The name of the dataclass.

    Returns:
        Dataclass definition.
    """
    dataclass_fields = []
    for field in typedef_type.fields:
        field_name = field.name
        field_name_to_use = f"{field_name}_" if field_name in kwlist else field_name
        dataclass_fields.append(
            (field_name_to_use, _idl_type_to_python_type(field.type, types)),
        )
    return _make_datacls(name, dataclass_fields)


_idl_enum_fields_named_to_dataclass_type_cache: dict[tuple[str, str], Type] = {}


def _idl_enum_fields_named_to_dataclass_type(
    fields: _IdlEnumFieldsNamed,
    types: Sequence[_IdlTypeDef],
    name: str,
) -> Type:
    dict_key = (name, str(fields))
    try:
        return _idl_enum_fields_named_to_dataclass_type_cache[dict_key]
    except KeyError:
        result = _idl_enum_fields_named_to_dataclass_type_no_cache(fields, types, name)
        _idl_enum_fields_named_to_dataclass_type_cache[dict_key] = result
        return result


def _idl_enum_fields_named_to_dataclass_type_no_cache(
    fields: _IdlEnumFieldsNamed,
    types: Sequence[_IdlTypeDef],
    name: str,
) -> Type:
    """Generate a dataclass definition from IDL named enum fields.

    Args:
        fields: The IDL enum fields.
        types: IDL type definitions.
        name: The name of the dataclass.

    Returns:
        Dataclass type definition.
    """
    dataclass_fields = []
    for field in fields:
        field_name = field.name
        field_name_to_use = f"{field_name}_" if field_name in kwlist else field_name
        dataclass_fields.append(
            (field_name_to_use, _idl_type_to_python_type(field.type, types)),
        )
    return _make_datacls(name, dataclass_fields)


def _idl_enum_fields_tuple_to_tuple_type(
    fields: _IdlEnumFieldsTuple,
    types: Sequence[_IdlTypeDef],
) -> Type:
    """Generate a tuple definition from IDL named enum fields.

    Args:
        fields: The IDL enum fields.
        types: IDL type definitions.

    Returns:
        Dataclass type definition.
    """
    tuple_types = []
    for field in fields:
        tuple_types.append(_idl_type_to_python_type(field, types))
    type_str = ",".join(t.__name__ for t in tuple_types)
    return eval(f"tuple[{type_str}]")  # noqa: WPS421,S307


def _idl_typedef_to_python_type(
    typedef: _IdlTypeDef,
    types: Sequence[_IdlTypeDef],
) -> Type:
    """Generate Python type from IDL user-defined type.

    Args:
        typedef: The user-defined type.
        types: IDL type definitions.

    Raises:
        ValueError: If an unknown type is passed.

    Returns:
        The Python type.
    """
    typedef_type = typedef.type
    if isinstance(typedef_type, _IdlTypeDefTyStruct):
        return _idl_typedef_ty_struct_to_dataclass_type(
            typedef_type,
            types,
            typedef.name,
        )
    elif isinstance(typedef_type, _IdlTypeDefTyEnum):
        return _handle_enum_variants(typedef_type, types, typedef.name).enum
    unknown_type = typedef_type.kind
    raise ValueError(f"Unknown type {unknown_type}")



"""Common utilities for encoding and decoding."""
from typing import Dict, Union
from hashlib import sha256

from anchorpy.idl import (
    Idl,
    _IdlEnumVariant,
    _IdlField,
    _IdlType,
    _IdlTypeDef,
    _IdlTypeDefTyEnum,
    _IdlTypeOption,
    _IdlTypeCOption,
    _IdlTypeArray,
    _IdlTypeDefined,
    _IdlTypeVec,
    _LiteralStrings,
    _NonLiteralIdlTypes,
)


def _sighash(ix_name: str) -> bytes:
    """Not technically sighash, since we don't include the arguments.

    (Because Rust doesn't allow function overloading.)

    Args:
        ix_name: The instruction name.

    Returns:
        The sighash bytes.
    """
    formatted_str = f"global:{ix_name}"
    return sha256(formatted_str.encode()).digest()[:8]


def _type_size_compound_type(idl: Idl, ty: _NonLiteralIdlTypes) -> int:
    if isinstance(ty, _IdlTypeVec):
        return 1
    if isinstance(ty, _IdlTypeOption):
        return 1 + _type_size(idl, ty.option)
    if isinstance(ty, _IdlTypeCOption):
        return 4 + _type_size(idl, ty.coption)
    if isinstance(ty, _IdlTypeDefined):
        defined = ty.defined
        filtered = [t for t in idl.types if t.name == defined]
        if len(filtered) != 1:
            raise ValueError(f"Type not found {ty}")
        type_def = filtered[0]
        return _account_size(idl, type_def)
    if isinstance(ty, _IdlTypeArray):
        element_type = ty.array[0]
        array_size = ty.array[1]
        return _type_size(idl, element_type) * array_size
    raise ValueError(f"type_size not implemented for {ty}")


def _type_size(idl: Idl, ty: _IdlType) -> int:
    """Return the size of the type in bytes.

    For variable length types, just return 1.
    Users should override this value in such cases.

    Args:
        idl: The parsed `Idl` object.
        ty: The type object from the IDL.

    Returns:
        The size of the object in bytes.
    """
    sizes: Dict[_LiteralStrings, int] = {
        "bool": 1,
        "u8": 1,
        "i8": 1,
        "bytes": 1,
        "string": 1,
        "i16": 2,
        "u16": 2,
        "u32": 4,
        "i32": 4,
        "u64": 8,
        "i64": 8,
        "u128": 16,
        "i128": 16,
        "publicKey": 32,
    }
    if isinstance(ty, str):
        return sizes[ty]
    return _type_size_compound_type(idl, ty)


def _variant_field_size(idl: Idl, field: Union[_IdlField, _IdlType]) -> int:
    if isinstance(field, _IdlField):
        return _type_size(idl, field.type)
    return _type_size(idl, field)


def _variant_size(idl: Idl, variant: _IdlEnumVariant) -> int:
    if variant.fields is None:
        return 0
    field_sizes = []
    field: Union[_IdlField, _IdlType]
    for field in variant.fields:
        field_sizes.append(_variant_field_size(idl, field))
    return sum(field_sizes)


def _account_size(idl: Idl, idl_account: _IdlTypeDef) -> int:
    """Calculate account size in bytes.

    Args:
        idl: The parsed `Idl` instance.
        idl_account: An item from `idl.accounts`.

    Returns:
        Account size.
    """
    idl_account_type = idl_account.type
    if isinstance(idl_account_type, _IdlTypeDefTyEnum):
        variant_sizes = (
            _variant_size(idl, variant) for variant in idl_account_type.variants
        )
        return max(variant_sizes) + 1
    if idl_account_type.fields is None:
        return 0
    return sum(_type_size(idl, f.type) for f in idl_account_type.fields)



"""Provides the Coder class."""
from anchorpy.coder.accounts import AccountsCoder
from anchorpy.coder.event import EventCoder
from anchorpy.coder.instruction import InstructionCoder
from anchorpy.idl import Idl


class Coder:
    """Coder provides a facade for encoding and decoding all IDL related objects."""

    def __init__(self, idl: Idl):
        """Initialize the coder.

        Args:
            idl: a parsed Idl instance.
        """
        self.instruction: InstructionCoder = InstructionCoder(idl)
        self.accounts: AccountsCoder = AccountsCoder(idl)
        self.events: EventCoder = EventCoder(idl)



"""This module contains code for creating the Anchor workspace."""
from typing import Optional, Union, cast, Dict
import json
from pathlib import Path
from solana.publickey import PublicKey
from anchorpy.program.core import Program
from anchorpy.provider import Provider
from anchorpy.idl import Idl, _Metadata

WorkspaceType = Dict[str, Program]


def create_workspace(
    path: Optional[Union[Path, str]] = None, url: Optional[str] = None
) -> WorkspaceType:
    """Get a workspace from the provided path to the project root.

    Args:
        path: The path to the project root. Defaults to the current working
            directory if omitted.
        url: The URL of the JSON RPC. Defaults to http://localhost:8899.

    Returns:
        Mapping of program name to Program object.
    """
    result = {}
    project_root = Path.cwd() if path is None else Path(path)
    idl_folder = project_root / "target/idl"
    for file in idl_folder.iterdir():
        with file.open() as f:
            idl_dict = json.load(f)
        idl = Idl.from_json(idl_dict)
        metadata = cast(_Metadata, idl.metadata)
        program = Program(idl, PublicKey(metadata.address), Provider.local(url))
        result[idl.name] = program
    return result


async def close_workspace(workspace: WorkspaceType) -> None:
    """Close the HTTP clients of all the programs in the workspace.

    Args:
        workspace: The workspace to close.
    """
    for program in workspace.values():
        # could do this in a faster way but there's probably no point.
        await program.close()



"""Extensions to the Borsh spec for Solana-specific types."""
from typing import Any, Dict, Type, TypeVar
from keyword import kwlist
from dataclasses import asdict
from borsh_construct import CStruct
from solana import publickey
from construct import Bytes, Adapter, Container


class _BorshPubkeyAdapter(Adapter):
    def __init__(self) -> None:
        super().__init__(Bytes(32))  # type: ignore

    def _decode(self, obj: bytes, context, path) -> publickey.PublicKey:
        return publickey.PublicKey(obj)

    def _encode(self, obj: publickey.PublicKey, context, path) -> bytes:
        return bytes(obj)


T = TypeVar("T")


class _DataclassStruct(Adapter):
    """Converts dataclasses to/from `borsh_construct.CStruct`."""

    def __init__(self, cstruct: CStruct, datacls: Type[T]) -> None:
        """Init.

        Args:
            cstruct: The underlying `CStruct`.
            datacls: The dataclass type.
        """
        super().__init__(cstruct)  # type: ignore
        self.datacls = datacls

    def _decode(self, obj: Container, context, path) -> T:
        kwargs = {}
        for key, value in obj.items():
            if key[0] != "_":
                key_to_use = f"{key}_" if key in kwlist else key
                kwargs[key_to_use] = value
        return self.datacls(**kwargs)  # type: ignore

    def _encode(self, obj: T, context, path) -> Dict[str, Any]:
        if isinstance(obj, dict):
            return obj
        return asdict(obj)


_BorshPubkey = _BorshPubkeyAdapter()  # noqa: WPS462
"""Adapter for (de)serializing a public key."""  # noqa: WPS322



"""This module contains code handling the Context object."""
from dataclasses import dataclass, field
from typing import Any, Tuple, Optional, Dict, List

from solana.keypair import Keypair
from solana.rpc.types import TxOpts
from solana.transaction import AccountMeta
from solana.transaction import TransactionInstruction

from anchorpy.idl import _IdlInstruction
from anchorpy.error import ArgsError


# should be Dict[str, Union[PublicKey, Accounts]]
# but mypy doesn't support recursive types
Accounts = Dict[str, Any]


@dataclass
class Context:
    """Context provides all non-argument inputs for generating Anchor transactions.

    Attributes:
        accounts: The accounts used in the instruction context.
        remaining_accounts: All accounts to pass into an instruction *after* the main
        `accounts`. This can be used for optional or otherwise unknown accounts.
        signers: Accounts that must sign a given transaction.
        pre_instructions: Instructions to run *before* a given method. Often this is
            used, for example to create accounts prior to executing a method.
        post_instructions: Instructions to run *after* a given method. Often this is
            used, for example to close accounts prior to executing a method.
        options: Commitment parameters to use for a transaction.

    """

    # For some reason mkdocstrings doesn't understand the full type hint
    # here if we use list[TransactionInstruction] instead of typing.List.
    # Weirdly there are other places where it understands list[whatever].

    accounts: Accounts = field(default_factory=dict)
    remaining_accounts: List[AccountMeta] = field(default_factory=list)
    signers: List[Keypair] = field(default_factory=list)
    pre_instructions: List[TransactionInstruction] = field(default_factory=list)
    post_instructions: List[TransactionInstruction] = field(default_factory=list)
    options: Optional[TxOpts] = None


def _check_args_length(
    idl_ix: _IdlInstruction,
    args: Tuple,
) -> None:
    """Check that the correct number of args is passed to the RPC function.

    Args:
        idl_ix: The IDL instruction object.
        args: The instruction arguments.

    Raises:
        ArgsError: If the correct number of args is not parsed.
    """
    if len(args) != len(idl_ix.args):
        expected_arg_names = [arg.name for arg in idl_ix.args]
        raise ArgsError(
            f"Provided incorrect number of args to instruction={idl_ix.name}. "
            f"Expected {expected_arg_names}",
            f"Received {args}",
        )


EMPTY_CONTEXT = Context()



"""This module contains utilities for the SPL Token Program."""
from typing import Optional
from solana.publickey import PublicKey
from solana.keypair import Keypair
from solana.rpc.types import RPCResponse
from solana.transaction import Transaction, TransactionInstruction
from solana.system_program import create_account, CreateAccountParams
from solana.utils.helpers import decode_byte_string
from spl.token.constants import TOKEN_PROGRAM_ID
from spl.token.instructions import (
    initialize_mint,
    InitializeMintParams,
    initialize_account,
    InitializeAccountParams,
    mint_to,
    MintToParams,
)
from spl.token.async_client import AsyncToken
from spl.token.core import AccountInfo, MintInfo
from spl.token._layouts import ACCOUNT_LAYOUT, MINT_LAYOUT
from anchorpy import Provider


async def create_token_account(
    prov: Provider,
    mint: PublicKey,
    owner: PublicKey,
) -> PublicKey:
    """Create a token account.

    Args:
        prov: An anchorpy Provider instance.
        mint: The pubkey of the token's mint.
        owner: User account that will own the new account.

    Returns:
        The pubkey of the new account.
    """
    token = AsyncToken(prov.connection, mint, TOKEN_PROGRAM_ID, prov.wallet.payer)
    return await token.create_account(owner)


async def create_token_account_instrs(
    provider: Provider,
    new_account_pubkey: PublicKey,
    mint: PublicKey,
    owner: PublicKey,
) -> tuple[TransactionInstruction, TransactionInstruction]:
    """Generate instructions for creating a token account.

    Args:
        provider: An anchorpy Provider instance.
        new_account_pubkey: The pubkey of the new account.
        mint: The pubkey of the token's mint.
        owner: User account that will own the new account.

    Returns:
        Transaction instructions to create the new account.
    """
    mbre_resp = await provider.connection.get_minimum_balance_for_rent_exemption(165)
    lamports = mbre_resp["result"]
    return (
        create_account(
            CreateAccountParams(
                from_pubkey=provider.wallet.public_key,
                new_account_pubkey=new_account_pubkey,
                space=165,
                lamports=lamports,
                program_id=TOKEN_PROGRAM_ID,
            )
        ),
        initialize_account(
            InitializeAccountParams(
                account=new_account_pubkey,
                mint=mint,
                owner=owner,
                program_id=TOKEN_PROGRAM_ID,
            )
        ),
    )


async def create_mint_and_vault(
    provider: Provider,
    amount: int,
    owner: Optional[PublicKey] = None,
    decimals: Optional[int] = None,
) -> tuple[PublicKey, PublicKey]:
    """Create a mint and a vault, then mint tokens to the vault.

    Args:
        provider: An anchorpy Provider instance.
        amount: The amount of tokens to mint to the vault.
        owner: User account that will own the new account.
        decimals: The number of decimal places for the token to support.

    Returns:
        The mint and vault pubkeys.
    """
    actual_owner = provider.wallet.public_key if owner is None else owner
    mint = Keypair()
    vault = Keypair()
    tx = Transaction()
    mint_space = 82
    create_mint_mbre_resp = (
        await provider.connection.get_minimum_balance_for_rent_exemption(mint_space)
    )
    create_mint_mbre = create_mint_mbre_resp["result"]
    create_mint_account_params = CreateAccountParams(
        from_pubkey=provider.wallet.public_key,
        new_account_pubkey=mint.public_key,
        space=mint_space,
        lamports=create_mint_mbre,
        program_id=TOKEN_PROGRAM_ID,
    )
    create_mint_account_instruction = create_account(
        create_mint_account_params,
    )
    init_mint_instruction = initialize_mint(
        InitializeMintParams(
            mint=mint.public_key,
            decimals=0 if decimals is None else decimals,
            mint_authority=provider.wallet.public_key,
            program_id=TOKEN_PROGRAM_ID,
        ),
    )
    vault_space = 165
    create_vault_mbre_resp = (
        await provider.connection.get_minimum_balance_for_rent_exemption(vault_space)
    )
    create_vault_mbre = create_vault_mbre_resp["result"]
    create_vault_account_instruction = create_account(
        CreateAccountParams(
            from_pubkey=provider.wallet.public_key,
            new_account_pubkey=vault.public_key,
            space=vault_space,
            lamports=create_vault_mbre,
            program_id=TOKEN_PROGRAM_ID,
        ),
    )
    init_vault_instruction = initialize_account(
        InitializeAccountParams(
            program_id=TOKEN_PROGRAM_ID,
            account=vault.public_key,
            mint=mint.public_key,
            owner=actual_owner,
        ),
    )
    mint_to_instruction = mint_to(
        MintToParams(
            program_id=TOKEN_PROGRAM_ID,
            mint=mint.public_key,
            dest=vault.public_key,
            amount=amount,
            mint_authority=provider.wallet.public_key,
        ),
    )
    tx.add(
        create_mint_account_instruction,
        init_mint_instruction,
        create_vault_account_instruction,
        init_vault_instruction,
        mint_to_instruction,
    )
    await provider.send(tx, [mint, vault])
    return mint.public_key, vault.public_key


def parse_token_account(info: RPCResponse) -> AccountInfo:
    """Parse `AccountInfo` from RPC response.

    Args:
        info: the `get_account_info` RPC response.

    Raises:
        ValueError: If the fetched data is the wrong size.
        AttributeError: If the account is not owned by the token program.

    Returns:
        The parsed `AccountInfo`.
    """
    if not info:
        raise ValueError("Invalid account owner")

    if info["result"]["value"]["owner"] != str(TOKEN_PROGRAM_ID):
        raise AttributeError("Invalid account owner")

    bytes_data = decode_byte_string(info["result"]["value"]["data"][0])
    if len(bytes_data) != ACCOUNT_LAYOUT.sizeof():
        raise ValueError("Invalid account size")

    decoded_data = ACCOUNT_LAYOUT.parse(bytes_data)

    mint = PublicKey(decoded_data.mint)
    owner = PublicKey(decoded_data.owner)
    amount = decoded_data.amount

    if decoded_data.delegate_option == 0:
        delegate = None
        delegated_amount = 0
    else:
        delegate = PublicKey(decoded_data.delegate)
        delegated_amount = decoded_data.delegated_amount

    is_initialized = decoded_data.state != 0
    is_frozen = decoded_data.state == 2

    if decoded_data.is_native_option == 1:
        rent_exempt_reserve = decoded_data.is_native
        is_native = True
    else:
        rent_exempt_reserve = None
        is_native = False

    if decoded_data.close_authority_option == 0:
        close_authority = None
    else:
        close_authority = PublicKey(decoded_data.owner)

    return AccountInfo(
        mint,
        owner,
        amount,
        delegate,
        delegated_amount,
        is_initialized,
        is_frozen,
        is_native,
        rent_exempt_reserve,
        close_authority,
    )


async def get_token_account(provider: Provider, addr: PublicKey) -> AccountInfo:
    """Retrieve token account information.

    Args:
        provider: The anchorpy Provider instance.
        addr: The pubkey of the token account.

     Returns:
        The parsed `AccountInfo` of the token account.
    """
    depositor_acc_info_raw = await provider.connection.get_account_info(addr)
    return parse_token_account(depositor_acc_info_raw)


async def get_mint_info(
    provider: Provider,
    addr: PublicKey,
) -> MintInfo:
    """Retrieve mint information.

    Args:
        provider: The anchorpy Provider instance.
        addr: The pubkey of the mint.

    Returns:
        The parsed `MintInfo`.
    """
    depositor_acc_info_raw = await provider.connection.get_account_info(addr)
    return parse_mint_account(depositor_acc_info_raw)


def parse_mint_account(info: RPCResponse) -> MintInfo:
    """Parse raw RPC response into `MintInfo`.

    Args:
        info: The RPC response from calling `.get_account_info` for the mint pubkey.

    Raises:
        AttributeError: If the account is not owned by the Token Program.
        ValueError: If the fetched data is the wrong size.

    Returns:
        The parsed `MintInfo`.
    """
    owner = info["result"]["value"]["owner"]
    if owner != str(TOKEN_PROGRAM_ID):
        raise AttributeError(f"Invalid mint owner: {owner}")

    bytes_data = decode_byte_string(info["result"]["value"]["data"][0])
    if len(bytes_data) != MINT_LAYOUT.sizeof():
        raise ValueError("Invalid mint size")

    decoded_data = MINT_LAYOUT.parse(bytes_data)
    decimals = decoded_data.decimals

    if decoded_data.mint_authority_option == 0:
        mint_authority = None
    else:
        mint_authority = PublicKey(decoded_data.mint_authority)

    supply = decoded_data.supply
    is_initialized = decoded_data.is_initialized != 0

    if decoded_data.freeze_authority_option == 0:
        freeze_authority = None
    else:
        freeze_authority = PublicKey(decoded_data.freeze_authority)

    return MintInfo(mint_authority, supply, decimals, is_initialized, freeze_authority)



"""The Python Anchor client."""
from anchorpy.provider import Provider, Wallet, SendTxRequest
from anchorpy.coder.coder import Coder, InstructionCoder, EventCoder, AccountsCoder
from anchorpy.idl import Idl, IdlProgramAccount
from anchorpy.workspace import create_workspace, close_workspace, WorkspaceType
from anchorpy.program.core import Program
from anchorpy.program.common import (
    Event,
    Instruction,
    translate_address,
    validate_accounts,
)
from anchorpy.program.context import Context
from anchorpy.program.namespace.account import AccountClient, ProgramAccount
from anchorpy.program.event import EventParser
from anchorpy.program.namespace.simulate import SimulateResponse
from anchorpy.pytest_plugin import localnet_fixture, workspace_fixture
from anchorpy import error, utils

__all__ = [
    "Program",
    "Provider",
    "Context",
    "create_workspace",
    "close_workspace",
    "Idl",
    "workspace_fixture",
    "WorkspaceType",
    "localnet_fixture",
    "Wallet",
    "SendTxRequest",
    "Coder",
    "InstructionCoder",
    "EventCoder",
    "AccountsCoder",
    "Instruction",
    "IdlProgramAccount",
    "Event",
    "translate_address",
    "validate_accounts",
    "AccountClient",
    "ProgramAccount",
    "EventParser",
    "SimulateResponse",
    "error",
    "utils",
]


__version__ = "0.8.2"
